function lume_update(t1, t2)
    for x, i in pairs(t2) do
        t1[x] = i
    end
    
    return t1
end

local importantItems = {
    potion_of_strength = 30,
    potion_of_life = 15,
    potion_of_healing = 40,
    potion_of_teleportation = 60,
    
    wand_of_blinking = 25,
    wand_of_teleportation = 30,
    wand_of_death = 5,
    wand_of_lightning = 10,
    wand_of_flames = 10,
    wand_of_switch_place = 10,
    wand_of_discord = 15,
    
    scroll_of_teleportation = 30,
    scroll_of_discord = 30,
    scroll_of_fear = 20,
    
    lore_note = 3,
    
    fadeleaf_seed = 2,
    flamesprout_seed = 2,
    flamesprout = 1
}


function spawnImportantItem(tile, chance)
    if math.random() <= (chance or .35) then
        return tile.room:spawnItem(lume.weightedchoice(importantItems), tile)
    end
    
    return false
end

function getImportantItem(chance)-- for i,x in pairs(importantItems) do assert(items[i], i) end
    if math.random() <= (chance or .35) then
        return lume.weightedchoice(importantItems)
    end
    
    return false
end

local weaponLevels = {0,1,2,2,3,-1,-2}

function getWeaponReward(level)
    return getWeapon(level or getValue(weaponLevels))
end

function spawnSkeletons(tile, room)
    local team = string.format("undead %s", math.random(1,3))
    return spawnGroup({skeleton=1}, {4,6}, function(cr, isLeader)
        -- add classes
        cr.team = team
        cr.hostility = 3
    end, tile)
end


function summonStatues(tile, room)
    return math.random() > .5 and room:spawnItem("angel_statue", tile) or spawnGroup({battle_statue=1}, {4,6}, function(cr, isLeader)
        -- add classes
        -- todo, link them, when one is attacked all is attacked
        room:addEnchantLight(cr.tile, 2, "cyan", .4)
    end, tile)
end


function juggernautSwarm(tile, room)
    return spawnGroup({juggernaut=1}, {6,8}, function(cr, isLeader)
        -- add classes
    end, tile)
end


function ratInfestation(tile, room)
    local crs = room:spawnFromData("goblin",tile.roomData,nil,{
            3, -- abritrary
            index = "juggernaut_beetle",
            intensity = 1, -- abritrary?
            group = {7,12},
            leader = false
    })
    return crs and crs[1]
end

function spawnWolfPack(tile, room)
    return spawnGroup({wolf=10,jackal=1}, {6,8}, function(cr, isLeader)
        -- add classes
    end, tile)
end

function spawnJackalPack(tile, room)
    return spawnGroup({jackal=1}, {2,3}, function(cr, isLeader)
        -- add classes
    end, tile)
end

function spawnBloats(tile, room)
    return spawnGroup({bloat=1}, {6,8}, function(cr, isLeader)
        -- add classes
    end, tile)
end

function spawnWispGroup(tile, room)
    return spawnGroup({wisp=1}, {6,8}, function(cr, isLeader)
        -- add classes
    end, tile)
end


function corruptionSpawn(tile, room)
    return true
end

function spawnDragon(tile, room)
    return tile.room:spawn("dragon", tile)
end

function spawnLich(tile, room)
    return true
end

function spawnBeholder(tile, room)
    return true
end

function spawnGoblinsVsOrcs(tile, room)
    if math.random() > .45 then
        return true
    end
    
    spawnGroup({goblin=1}, {2,4}, function(cr, isLeader)
        -- add classes
        -- cr.team = team
        cr.hostility = 3
    end, tile)
    spawnGroup({orc=1}, {2,4}, function(cr, isLeader)
        -- add classes
        -- cr.team = team
        cr.hostility = 3
    end, tile)
end

--[[
spawnImportantItem(tile)
spawnSkeletons(tile, map)
summonStatues
juggernautSwarm
ratInfestation
spawnDragon
spawnWispGroup
spawnBloats
corruptionSpawn
spawnLich
spawnBeholder
spawnWolfPack
spawnJackalPack
]]


PIG_DATA = {
    structures = {
        goblin_warren = {
            action = buildGoblinWarren,
        },
        garden = {
            action = null
        },
        kobold_camp = {
            action = buildKoboldCamp,
        },
        goblin_camp = {
            action = buildGoblinCamp,
        },
        juggernaut_nest = {
            action = buildJuggernautNest,
        },
        labyrinth = {
            action = buildLabyrinth,
        },
        bat_cave  = {
            action = buildBatCave
        },
        battleStatueItem = {
            action = buildBattleStatueItem,
        },
        chasmItem = {
            action = buildChasmItem,
        },
        itemGraveRoom = {
            action = buildItemGraveRoom,
        }
    },
    
    events = {
        -- events have their action function played on an unseen tile
            
        marchOfTheDead = {
            action = spawnSkeletons,
            text = {
                basic = {
                    "You feel a cold chill run down your spine.",
                    "The air feels colder...",
                }
            }
        },
        summonStatues = {
            action = summonStatues,
            text = {
                basic = {
                    "You hear a loud @Thud!",
                }
            }
        },
        juggernautSwarm = {
            action = juggernautSwarm,
            text = {
                basic = {
                    "You hear... buzzing?"
                }
            }
        },
        ratInfestation = {
            action = ratInfestation,
            text = {
                basic = {
                    "smells...wierd"
                }
            }
        },
        dragonSpawn = {
             action = spawnDragon,
             text = {
                 basic = {
                    "The air seems hotter. Waaay hotter.",
                    "Your sweat dries up.",
                    "The heat is suffocating...",
                }
            }
        },
        wispGroup = {
            action = spawnWispGroup,
            text = {
                basic = {
                    "The air seems warmer...",
                }
            }
        },
        bloatSpawn = {
            action = spawnBloats,
            text = {
                basic = "You find it harder to breathe."
            }
        },
        corruption = {
            action = corruptionSpawn,
            text = {
                basic = "No....no...",
            }
        },
        wanderingLich = {
            action = spawnLich,
            text = {
                basic = {
                    "Magic's in the air?",
                    "The air seems ...more evil."
                }
            }
        },
        beholderGroup = {
            action = spawnBeholder,
            text = {
                basic = {
                    "You feel watched.",
                }
            }
        },
        wolfPack = {
            action = spawnWolfPack,
            text = {
                basic = {
                    "You feel....delicious?",
                }
            }
        },
        jackalPack = {
            action = spawnJackalPack,
            -- text = nil
        },
        goblinVOrcBattle = {
            action = spawnGoblinsVsOrcs,
            text = {
                basic = {
                    "Ares seems to be laughing.",
                    "The dungeon seems to be filled with anger.",
                },
            }
        },
        koboldCamp = {
            action = buildKoboldCamp,
            text = {
                basic = {
                    "Mischief settles...",
                },
            },
        },
        goblinCamp = {
            action = buildGoblinCamp,
            text = {
                basic = {
                    "The dungeon makes space for mischief",
                },
            }
        }
    
    },
        
    MAP_SIZES = {
        vlarge = {120, 120},
        big = {80, 80},
        medium = {40, 40},
        small = {20, 30}
    },
    
    SPECIES = {
        pig = function(prep)
        
            --[[ocal check = {"These structures and events are missing an action function"}
            for x, i in pairs(PIG_DATA.structures) do
                if not i.action then
                    check[2] = true
                    check["structure: "..x] = true
                end
            end
            
            for x, i in pairs(PIG_DATA.events) do
                if not i.action then
                    check[2] = true
                    check["event: "..x] = true
                end
            end]]
            
            -- assert(not check[2], inspect(check))
            
            local dat = lume_update(lume.copy(creatures.pig), {
                eatRate = 1.75,
                inventorySpace = lume.max(creatures.pig.inventorySpace or 0, 8)+3,
                inventory = prep == 0 and {} or {
                    {"carrot", prep+1},
                    {"potion_of_strength", prep == 1 and 1 or {prep, prep+3}},
                    {"potion_of_healing", prep == 1 and {1,3} or {prep, prep+3}},
                    {"potion_of_levitation", 1},

                    {{"wand_of_flames", "wand_of_lightning", "wand_of_flames" or "wand_of_entrancement", "wand_of_webbing"}, {prep-1, prep+1}}
                }
            })
            
            local space = 5+3
            for index, value in ipairs(dat.inventory) do
                local num = (type(value) == "table" and getValue(value[2])) 
                space = space + (type(num) == "number" and num or 1)
            end
            
            dat.inventorySpace = lume.max(dat.inventorySpace, space)
            dat.drop = { meat = {30, {1,1,1,1,2},"meat"}}
            dat.isPig = true
            
            return dat
        end
    },
    
    ECOLOGY = {
        [0] = {
            -- basic
            spawn = {
                goblin          = {weight = 15},
                kobold          = {weight = 25},
                jackal          = {weight = 30},
                pig             = {weight = 50},
                chicken         = {weight = 30},
                bat             = {weight = 10},
                wolf            = {weight = 3},
                bear            = {weight = 3},
                ogre            = {weight = 2},
                rat             = {weight = 15},
                toad            = {weight = 10},
                spider          = {weight = 5},
                snail           = {weight = 0},
                wisp            = {weight = 0},
                bloat           = {weight = 0},
                giant_snail     = {weight = 0},
                golem           = {weight = 0},
                dragon          = {weight = 0},
                angel           = {weight = 1 or 0},
                orcs            = {weight = 0},
                juggernaut      = {weight = 0},
                gelatinous_cube = {weight = 0},
            },
            
            structures = {
                goblin_warren    = {weight = 30},
                garden           = {weight = 30},
                kobold_camp      = {weight = 20},
                labyrinth        = {weight = 5},
                juggernaut_nest  = {weight = 15},
                goblin_camp      = {weight = 50},
                bat_cave         = {weight = 40},
                battleStatueItem = {weight = 10},
                chasmItem        = {weight = 40},
                itemGraveRoom    = {weight = 6}
            },
            
            events = {
                -- an event is picked based on weight after time
                marchOfTheDead   = {weight = 5},
                summonStatues    = {weight = 2},
                juggernautSwarm  = {weight = 2},
                ratInfestation   = {weight = 3},
                dragonSpawn      = {weight = 1},
                wispGroup        = {weight = 4},
                bloatSpawn       = {weight = 1},
                corruption       = {weight = 1},
                wanderingLich    = {weight = 1},
                
                -- spawn beholder in a room/premade structure
                beholderGroup    = {weight = 1},
                
                wolfPack         = {weight = 1},
                goblinVOrcBattle = {weight = 1},
                koboldCamp       = {weight = 1},
                goblinCamp       = {weight = 1}
            },
            
            eventTimers = {
                often = {weight = 5, timer = {3, 7}},
                periodic = {weight = 7, timer = {10, 15}},
                rare = {weight = 5, timer = {20, 30}},
                scanty = {weight = 2, timer = {30, 40}}
            }
        } -- spawn table of some kind, including structures and timed events
    },
    
    PREPARATION = {[0] = true, true, nil, nil}
}

for x, i in pairs(PIG_DATA.events) do
    if not i.action then error(x) end
end


--[[
    Preparation
        0, 1, 2, 3
    Ecology
        0, 1, 2, 3
    Species
        pig, iackal, kobold, elf
    Map
        vlarge, big, medium, small
]]